/* main.c
 * Copyright 2021 Steven Arnow <s@rdw.se>
 * 
 * This file is part of keyboard-dummy.
 * 
 * keyboard-dummy is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * keyboard-dummy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with keyboard-dummy.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdint.h>

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>

#include "usbdrv.h"

#include <util/delay.h>

#define LOOKUP_IN_EEPROM

#include "keycodes.h"

#ifdef LOOKUP_IN_EEPROM
#include <avr/eeprom.h>
#endif

// From Frank Zhao's USB Business Card project
// http://www.frank-zhao.com/cache/usbbusinesscard_details.php
// Modified by Steven Arnow for media keys
PROGMEM const char usbHidReportDescriptor[USB_CFG_HID_REPORT_DESCRIPTOR_LENGTH] = {
    0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
    0x09, 0x06,                    // USAGE (Keyboard)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x75, 0x01,                    //   REPORT_SIZE (1)
    0x95, 0x08,                    //   REPORT_COUNT (8)
    0x05, 0x07,                    //   USAGE_PAGE (Keyboard)(Key Codes)
    0x19, 0xe0,                    //   USAGE_MINIMUM (Keyboard LeftControl)(224)
    0x29, 0xe7,                    //   USAGE_MAXIMUM (Keyboard Right GUI)(231)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x25, 0x01,                    //   LOGICAL_MAXIMUM (1)
    0x81, 0x02,                    //   INPUT (Data,Var,Abs) ; Modifier byte
    0x95, 0x01,                    //   REPORT_COUNT (1)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x81, 0x03,                    //   INPUT (Cnst,Var,Abs) ; Reserved byte
    0x95, 0x05,                    //   REPORT_COUNT (5)
    0x75, 0x01,                    //   REPORT_SIZE (1)
    0x05, 0x08,                    //   USAGE_PAGE (LEDs)
    0x19, 0x01,                    //   USAGE_MINIMUM (Num Lock)
    0x29, 0x05,                    //   USAGE_MAXIMUM (Kana)
    0x91, 0x02,                    //   OUTPUT (Data,Var,Abs) ; LED report
    0x95, 0x01,                    //   REPORT_COUNT (1)
    0x75, 0x03,                    //   REPORT_SIZE (3)
    0x91, 0x03,                    //   OUTPUT (Cnst,Var,Abs) ; LED report padding
    0x95, 0x06,                    //   REPORT_COUNT (6)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x25, 0xEB,                    //   LOGICAL_MAXIMUM (235)
    0x05, 0x07,                    //   USAGE_PAGE (Keyboard)(Key Codes)
    0x19, 0x00,                    //   USAGE_MINIMUM (Reserved (no event indicated))(0)
    0x29, 0xEB,                    //   USAGE_MAXIMUM (Keyboard Media Next)(235)
    0x81, 0x00,                    //   INPUT (Data,Ary,Abs)
    0xc0                           // END_COLLECTION
};

#ifndef LOOKUP_IN_EEPROM

PROGMEM const uint8_t g_keycode_lookup[88] = {
	0
};

PROGMEM const uint8_t g_keycode_lookup2[15] = {
	0
};

#endif


typedef struct {
	uint8_t modifier;
	uint8_t reserved;
	uint8_t keycode[6];
} keyboard_report_t;

static keyboard_report_t g_krp;
static uchar g_idle_rate;
static uchar g_led_state = 0xff;


USB_PUBLIC uchar usbFunctionSetup(uchar data[8]) {
	usbRequest_t *rq = (void *) data;

	if ((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS) {
		switch (rq->bRequest) {
			case USBRQ_HID_GET_REPORT:
				usbMsgPtr = (void *) &g_krp;
				g_krp.modifier = 0;
				g_krp.keycode[0] = 0;
				return sizeof(g_krp);
			case USBRQ_HID_SET_REPORT:
				return (rq->wLength.word == 1) ? USB_NO_MSG : 0;
			case USBRQ_HID_GET_IDLE:
				usbMsgPtr = &g_idle_rate;
				return 1;
			case USBRQ_HID_SET_IDLE:
				g_idle_rate = rq->wValue.bytes[1];
				return 0;
		}
	}

	return 0;
}


usbMsgLen_t usbFunctionWrite(uint8_t *data, uchar len) {
	if (g_led_state == data[0])
		return 1;
	g_led_state = data[0];
	PORTB &= (~0x7);
	PORTB |= ((~g_led_state) & 0x7);

	return 1;
}


int main() {

	// Initialize USI for SPI
	//USICR = 0x10; // 3-wire mode, software clock
	USICR = 0;

	// Set outputs
	DDRB = (1 << 7) | (1 << 6) | (1 << 2) | (1 << 1) | (1 << 0);
	DDRD = (1 << 0) | (1 << 1);




	wdt_enable(WDTO_1S);
	usbInit();
	// I used to have a reset thing here that poked the pull-up resistor.
	// But I ran out of code space...

	sei();


	for (;;) {
		wdt_reset();
		usbPoll();

		if (usbInterruptIsReady() && g_led_state != 0xff) {
			/* Make report */ { 
				uchar i, j;
				PORTD &= (~0x3); // Lower ~LOAD, ~CS
				
				for (i = 0; i < 6; i++)
					g_krp.keycode[i] = 0;
				g_krp.modifier = 0;
				
				PORTD |= 0x2; // Raise ~LOAD again
				
//				for (i = j = 0; i < 6 && j < 96; j++) {
				for (i = j = 0; i < 6 && j < 16; j++) {
					uint8_t data;

					// Toggle the clock twice...
					PORTB |= 0x80;

					data = !!(PINB & (1 << 5));
					PORTB &= ~(1 << 6);
					PORTB |= (data << 6);
					
					PORTB ^= 0x80;
				
					if (!data)
						continue;

					if (j < 8) { // Check modifier
						g_krp.modifier |= (data << j);
					} else {
						data = j - 8;
						//g_krp.keycode[i++] = USB_KBD_KEY_A + j;

						if (j < 32 && g_led_state & 1) {
							#ifdef LOOKUP_IN_EEPROM
							data += 96; // Right after normal table
							data = eeprom_read_byte((uint8_t *) data);
							#else
							data = g_keycode_lookup2[data];
							#endif
						} else {
							#ifdef LOOKUP_IN_EEPROM
							data = eeprom_read_byte((uint8_t *) data);
							#else
							data = g_keycode_lookup[data];
							#endif
						}
						g_krp.keycode[i++] = data;
					}

					

				}

				PORTD |= 0x1; // Raise ~CS again
			}

			usbSetInterrupt((void *) &g_krp, sizeof(g_krp));
		}

	}
	
	return 0;
}
